/**
 * This class is used to create a Client object.
 * The client handels the client server communication.
 * It communicates with the server (e.g. connection or query)
 * and recieves data from it, if dynamicMode is true.
 * If dynamicMode is false, the client can also handle data that is already
 * loaded and is not coming in in chunks.
 */
class Client {
  constructor({corpusId = null,
               socket = null,
               logging = true,
               dynamicMode = true} = {}) {
    this.corpusId = corpusId;
    this.dynamicMode = dynamicMode;
    this.logging = logging;
    this.requestQueryProgress = 0;
    this.socket = socket;
    this.socketEventListeners = {};
    this.recivedMetaData = false;

    /**
     * Disable all console logging.
     * This is global. So every other log message in every other Class or
     * function used in conjunction with the client either logs or does not
     * log depending on the logging flag.
     * Credits to https://gist.github.com/kmonsoor/0244fdb4ad79a4826371e58a1a5fa984
     */
    if (!logging) {
      (() => {
        let console = (window.console = window.console || {});
        [
          'assert', 'clear', 'count', 'debug', 'dir', 'dirxml',
          'error', 'exception', 'group', 'groupCollapsed', 'groupEnd',
          'info', 'log', 'markTimeline', 'profile', 'profileEnd', 'table',
          'time', 'timeEnd', 'timeStamp', 'trace', 'warn'
        ].forEach((method) => {
          console[method] = () => {};
        });
      })();
    }
    console.info("Client initialized:", this);
  }

  // Registers one or more SocketEventListeners to the Client.
  setSocketEventListeners(socketEventListeners) {
    for (let socketEventListener of socketEventListeners) {
      this.socketEventListeners[socketEventListener.type] = socketEventListener;
    }
  }

  /**
   * Loads the SocketEventListeners so they will be triggered on their assigned
   * type strings because they double as the socket event event names.
   */
  loadSocketEventListeners() {
    for (let [type, listener] of Object.entries(this.socketEventListeners)) {
      listener.listenerFunction(type, this);
    }
  }

  /**
   * This functions sends events to the View to trigger specific functions that
   * are handleing the represnetation of data stored in the model.
   */
  notifyView(SendWhatToDo) {

  }

  // Registers a CorpusAnalysisDisplay object to the Client.
  setDisplay(type, corpusAnalysisDisplay) {
    this.displays[type] = corpusAnalysisDisplay;
  }

  /**
   * Function that takes one or more query selector
   * strings in an array as an input. The function then creates a
   * class field in the client object with the query selector
   * string as the key. The selector will be converted to a valid JavaScript
   * Field name i. e. #html-id-string -> this.htmlIdString
   * The value will be the identifed element fetched with the querySelector
   * method.
   */
  // TODO: multipleResults=false, atattchSomeCallback=false ?
  getHTMLElements(arrayOfSelectors) {
    for (let selector of arrayOfSelectors) {
      let element = document.querySelector(selector);
      let cleanKey = [];
      selector = selector.replace('_', '-');
      selector.match(/\w+/g).forEach((word) => {
        let tmp = word[0].toUpperCase() + word.slice(1);
        cleanKey.push(tmp);
      });
      cleanKey[0] = cleanKey[0].toLowerCase();
      cleanKey = cleanKey.join('');
      this[cleanKey] = element;
    }
  }

  /**
   * Connects to the corpus analysis session for the specified corpus via
   * socket.io.
   */
  connect() {
    console.info('corpus_analysis_init: Client connecting to session via',
                 'socket.emit');
    // if (this.displays.init != undefined) {
    //   this.displays.init.element.M_Modal.open();
    //   this.displays.init.setVisibilityByStatus('waiting');
    // }
    this.socket.emit('corpus_analysis_init', this.corpusId);
  }

  getMetaData() {
    console.info('corpus_analysis_meta_data: Client getting meta data via',
                 'socket.emit.');
    this.socket.emit('corpus_analysis_meta_data', this.corpusId);
  }

  /**
   * Emits query to the server via socket.io. Server will send the results
   * back.
   */
  query(queryStr) {
    console.info('corpus_analysis_query: Client sending query via',
                 'socket.emit for the query', queryStr);
    this.socket.emit('corpus_analysis_query', queryStr);
  }
}

/**
 * This class is used to create an SocketEventListener.
 * Input are an identifying type string, the listener function and callbacks
 * which will be executed as part of the listener function. The identifying
 * type string is also used as the socket event event identifier.
 */
class SocketEventListener {
  constructor(type, listenerFunction, args=null) {
    this.listenerCallbacks = {};
    this.listenerFunction = listenerFunction;
    this.type = type;
  }

  // Registers callbacks to this SocketEventListener
  setCallbacks(listenerCallbacks) {
    for (let listenerCallback of listenerCallbacks) {
      this.listenerCallbacks[listenerCallback.type] = listenerCallback;
    }
  }

  /** Shorthand to execute all registered callbacks with same args in insertion
   * order.
   * NOTE:
   * Since ECMAScript 2015, objects do preserve creation order for
   * string and Symbol keys. In JavaScript engines that comply with the
   * ECMAScript 2015 spec, iterating over an object with only string keys will
   * yield the keys in order of insertion.
   * So all modern Browsers.
   */
  executeCallbacks(payload) {
    for (let [type, listenerCallback] of Object.entries(this.listenerCallbacks)) {
      listenerCallback.args.unshift(payload);
      listenerCallback.callbackFunction(...listenerCallback.args);
    }
  }
}

/**
 * This class is used to create an ListenerCallback which will be registered
 * to an SocketEventListener so the Listener cann invoke the ListenerCallback
 * callback functions.
 */
class ListenerCallback {
  constructor(type, callbackFunction, argsList) {
    this.args = argsList;
    this.callbackFunction = callbackFunction;
    this.type = type;
  }
}

// export Classes from this module
export {
  Client,
  SocketEventListener,
  ListenerCallback,
};